<?php
/**
 * Set the content width in pixels, based on the theme's design and stylesheet.
 *
 * Priority 0 to make it available to lower priority callbacks.
 *
 * @global int $content_width
 */
function gutenbee_content_width() {
	$content_width = $GLOBALS['content_width'];

	$info          = gutenbee_get_layout_info();
	$content_width = $info['content_width'];

	$GLOBALS['content_width'] = apply_filters( 'gutenbee_content_width', $content_width );
}
add_action( 'template_redirect', 'gutenbee_content_width', 0 );

/**
 * Returns the appropriate html classes for a particular number of grid columns.
 *
 * @param int $columns A number of grid columns.
 *
 * @return string
 */
function gutenbee_get_theme_columns_classes( $columns ) {
	switch ( intval( $columns ) ) {
		case 1:
			$classes = 'col-12';
			break;
		case 2:
			$classes = 'col-md-6 col-12';
			break;
		case 3:
			$classes = 'col-lg-4 col-md-6 col-12';
			break;
		case 4:
		default:
			$classes = 'col-xl-3 col-md-6 col-12';
			break;
	}

	return apply_filters( 'gutenbee_get_theme_columns_classes', $classes, $columns );
}

/**
 * Returns the appropriate html classes for a particular number of grid columns (specifically for the blog listing).
 *
 * @see gutenbee_get_theme_columns_classes()
 *
 * @param int $columns A number of grid columns.
 *
 * @return string
 */
function gutenbee_get_blog_columns_classes( $columns ) {
	switch ( intval( $columns ) ) {
		case 1:
			$classes = 'col-12';
			break;
		case 2:
			$classes = 'col-xl-6 col-12';
			break;
		case 3:
			$classes = 'col-xl-4 col-lg-6 col-12';
			break;
		case 4:
		default:
			$classes = 'col-xl-3 col-lg-6 col-12';
			break;
	}

	return apply_filters( 'gutenbee_get_blog_columns_classes', $classes, $columns );
}

/**
 * Determine if a sidebar can be displayed.
 *
 * This function determines if a sidebar *can* be displayed, not if it *will* be displayed.
 * For example, the pages' sidebar *can* be displayed if there are widgets assigned onto it,
 * but it may not be displayed (decided elsewhere), depending on user settings or other factors.
 *
 * @return bool
 */
function gutenbee_has_sidebar() {
	$has_sidebar = false;

	if ( gutenbee_is_woocommerce_with_sidebar() ) {
		$has_sidebar = is_active_sidebar( 'shop' );
	} elseif ( gutenbee_is_woocommerce_without_sidebar() ) {
		$has_sidebar = false;
	} elseif ( ! is_page() ) {
		$has_sidebar = is_active_sidebar( 'sidebar-1' );
	} elseif ( is_page() ) {
		$has_sidebar = is_active_sidebar( 'sidebar-2' );
	}

	return apply_filters( 'gutenbee_has_sidebar', $has_sidebar );
}

/**
 * Return the layout name of the current page.
 *
 * @return string
 */
function gutenbee_get_layout_info_layout() {
	$layout = '';

	if ( is_home() || is_singular( 'post' ) || is_category() || is_tag() || is_author() || is_date() || is_search() ) {
		$layout = gutenbee_get_mod( 'blog_layout_archives_layout_type', gutenbee_blog_layout_archives_layout_type_default() );
	} elseif ( gutenbee_is_woocommerce_without_sidebar() ) {
		$layout = 'fullwidth';
	} else {
		$layout = gutenbee_get_mod( 'global_layout_pages_layout_type', gutenbee_global_layout_pages_layout_type_default() );
	}

	return apply_filters( 'gutenbee_get_layout_info_layout', $layout );
}

/**
 * Return appropriate layout information for a specific layout.
 *
 * @uses gutenbee_get_layout_info_layout()
 *
 * @param false|string $layout The layout name to get information for, or empty to auto-detect.
 *
 * @return array
 */
function gutenbee_get_layout_info( $layout = false ) {
	$has_sidebar = gutenbee_has_sidebar();

	$defaults = gutenbee_customizer_defaults();

	// Assume a default layout of "Content / Sidebar", and then handle all other cases.

	$content_width = (array) get_theme_mod( 'global_layout_content_width', $defaults['global_layout_content_width'] );
	$content_width = ! empty( $content_width['desktop'] ) ? $content_width['desktop'] : $defaults['global_layout_content_width']['desktop'];
	$sidebar_width = (array) get_theme_mod( 'global_layout_sidebar_width', $defaults['global_layout_sidebar_width'] );
	$sidebar_width = ! empty( $sidebar_width['desktop'] ) ? $sidebar_width['desktop'] : $defaults['global_layout_sidebar_width']['desktop'];

	$classes = array(
		'layout'            => $layout ? $layout : gutenbee_get_layout_info_layout(),
		'container_classes' => "col-lg-{$content_width} col-12",
		'sidebar_classes'   => $has_sidebar ? "col-lg-{$sidebar_width} col-12" : '',
		'content_width'     => 960,
		'has_sidebar'       => $has_sidebar,
		'row_classes'       => $has_sidebar ? 'has-sidebar' : 'justify-content-center',
	);


	switch ( $classes['layout'] ) {
		case 'sidebar_content':
			// If there's no sidebar to show, we want to keep the .justify-content-center class.
			if ( $has_sidebar ) {
				$classes['row_classes'] = 'has-sidebar flex-row-reverse';
			}
			break;
		case 'fullwidth':
			$classes = array(
				'layout'            => 'fullwidth',
				'container_classes' => 'col-12',
				'sidebar_classes'   => '',
				'content_width'     => 1260,
				'has_sidebar'       => false,
				'row_classes'       => '',
			);
			break;
		case 'fullwidth_narrow':
			$classes['sidebar_classes'] = '';
			$classes['has_sidebar']     = false;
			$classes['row_classes']     = 'justify-content-center';
			break;
	}

	if ( 0 === strpos( $classes['layout'], 'listing_' ) && ! $has_sidebar ) {
		$classes['container_classes'] = 'col-12';
	}

	return apply_filters( 'gutenbee_layout_info', $classes, $has_sidebar );
}
